from typing import Iterable

from eric_sse.persistence import KvStorageEngine, ItemNotFound
from eric_sse.repository import ChannelRepository
from eric_sse.entities import AbstractChannel


class InMemoryStorage(KvStorageEngine):

    def __init__(self, objects: dict[str, any] = None):
        self.objects = objects or {}

    objects: dict[str, any] = {}

    def fetch_all(self) -> Iterable[any]:
        for obj in self.objects.values():
            yield obj

    def upsert(self, key: str, value: any):
        self.objects[key] = value

    def fetch_one(self, key: str) -> any:
        try:
            return self.objects[key]
        except KeyError:
            raise ItemNotFound(key)

    def delete(self, key: str):
        del InMemoryStorage.objects[key]

class InMemoryChannelRepository(ChannelRepository):
    def __init__(self, channels: dict[str, AbstractChannel] = None):
        super().__init__(storage_engine=InMemoryStorage(objects=channels or {}))




import eric_sse
from eric_sse.exception import RepositoryError
from eric_sse.message import MessageContract
from eric_sse.persistence import ObjectAsKeyValuePersistenceMixin

logger = eric_sse.get_logger()

class MessageQueueListener:
    """
    Base class for listeners.

    Optionally you can override on_message method if you need to inject code at message delivery time.
    """

    def __init__(self):
        self.id: str = eric_sse.generate_uuid()
        self.__is_running: bool = False

    def on_message(self, msg: MessageContract) -> None:
        """Event handler. It executes when a message is delivered to client"""
        pass

    def start(self) -> None:
        self.__is_running = True

    def stop(self) -> None:
        self.__is_running = False

    def is_running(self) -> bool:
        return self.__is_running


class PersistableListener(MessageQueueListener, ObjectAsKeyValuePersistenceMixin):
    """Gives KV persistence support to MessageQueueListener."""
    @property
    def kv_key(self) -> str:
        return self.id

    @property
    def kv_setup_values_as_dict(self) -> dict:
        return {
            'id': self.id,
            'is_running': self.is_running()
        }

    def kv_setup_by_dict(self, setup: dict):
        try:
            self.id = setup['id']
            if setup.get('is_running', False):
                self.start()
        except KeyError:
            raise RepositoryError(f'Invalid setup: {setup}')

    @property
    def kv_constructor_params_as_dict(self) -> dict:
        return {}



from abc import ABC, abstractmethod
from typing import Iterable

from eric_sse.entities import AbstractChannel
from eric_sse.connection import Connection
from eric_sse.queues import Queue
from eric_sse.listener import MessageQueueListener


class ChannelRepositoryInterface(ABC):

    @abstractmethod
    def load_all(self) -> Iterable[AbstractChannel]:
        pass

    @abstractmethod
    def load_one(self, key: str) -> AbstractChannel:
        pass

    @abstractmethod
    def persist(self, persistable: AbstractChannel):
        pass

    @abstractmethod
    def delete(self, key: str):
        pass

class ConnectionRepositoryInterface(ABC):

    @abstractmethod
    def load_all(self, channel_id: str) -> Iterable[Connection]:
        pass

    @abstractmethod
    def load_one(self, channel_id: str, key: str) -> Connection:
        pass

    @abstractmethod
    def persist(self, channel_id: str, connection: Connection):
        pass

    @abstractmethod
    def delete(self, key: str):
        pass

class ListenerRepositoryInterface(ABC):

    @abstractmethod
    def load_all(self) -> Iterable[MessageQueueListener]:
        pass

    @abstractmethod
    def load_one(self, channel_id: str, key: str) -> MessageQueueListener:
        pass

    @abstractmethod
    def persist(self, channel_id: str, listener: MessageQueueListener):
        pass

    @abstractmethod
    def delete(self, key: str):
        pass

class QueueRepositoryInterface(ABC):

    @abstractmethod
    def load_one(self, key: str) -> Queue:
        pass

    @abstractmethod
    def persist(self, queue: Queue):
        pass

    @abstractmethod
    def delete(self, key: str):
        pass


"""
This module is intended to those who want to create their own persistence layer

A **Redis** concrete implementation of interfaces is available at  `eric-redis-queues package source <https://github.com/laxertu/eric-redis-queues/blob/master/eric_redis_queues/__init__.py>`_.

Here is a `Basic example <https://github.com/laxertu/eric/blob/master/examples/use_cases/basic.py>`_ of integration.

**Writing a custom persistence layer**

You'll need to implement the following interfaces:

**Channels**

* :class:`~eric_sse.persistence.ChannelRepositoryInterface`

* You'll need to define a channel that implements :class:`~eric_sse.persistence.PersistableChannel` if :class:`~eric_sse.prefabs.SSEChannel` do not suit with your requirements
* For **MessageQueueListener** support you can extend or directly use :class:`~eric_sse.persistence.PersistableListener`.

**Connections**

* :class:`~eric_sse.persistence.PersistableQueue`
* :class:`~eric_sse.persistence.ConnectionRepositoryInterface`
"""

from abc import ABC, abstractmethod
from typing import Iterable
from importlib import import_module

from eric_sse.exception import RepositoryError

class ItemNotFound(Exception):
    pass



class ObjectAsKeyValuePersistenceMixin(ABC):
    """
    Adds KV persistence support.

    By implementing this abstract mixin should be possible to persist every object that is not directly
    serializable by pickle, for example, if your Queues implementation wraps some incompatible dependency, e.g. a Redis client.

    For this reason, the idea is that dict values should be serializable by pickle too.

    see :func:`~eric_sse.persistence.importlib_create_instance`
    """
    @property
    def dict(self):
        return {
            'class_path': self.kv_class_absolute_path,
            'constructor_parameters': self.kv_constructor_params_as_dict,
            'setup_dict': self.kv_setup_values_as_dict
        }

    @property
    @abstractmethod
    def kv_key(self) -> str:
        """The key to use when persisting object"""
        ...

    @property
    @abstractmethod
    def kv_setup_values_as_dict(self) -> dict:
        """Returns value that will be persisted as a dictionary."""
        ...

    @abstractmethod
    def kv_setup_by_dict(self, setup: dict):
        """Does necessary post-creation setup of object given its persisted values"""
        ...

    @property
    def kv_class_absolute_path(self) -> str:
        """Returns class full path as string"""
        return f'{self.__module__}.{type(self).__name__}'

    @property
    @abstractmethod
    def kv_constructor_params_as_dict(self) -> dict:
        """Class constructor parameters as dict"""
        ...


class KvStorageEngine(ABC):
    @abstractmethod
    def fetch_all(self) -> Iterable[any]:
        ...

    @abstractmethod
    def upsert(self, key: str, value: any):
        ...

    @abstractmethod
    def fetch_one(self, key: str) -> any:
        ...

    @abstractmethod
    def delete(self, key: str):
        ...


class ObjectRepositoryInterface(ABC):
    """Every exception raised by concrete implementations show be wrapped inside a :class:`~eric_sse.exception.RepositoryError`
    """

    @abstractmethod
    def load_all(self) -> Iterable[ObjectAsKeyValuePersistenceMixin]:
        """Returns an Iterable of all persisted objects of correspondant concrete implementation."""
        ...

    @abstractmethod
    def load_one(self, key: str) -> ObjectAsKeyValuePersistenceMixin:
        ...

    @abstractmethod
    def persist(self, persistable: ObjectAsKeyValuePersistenceMixin):
        ...

    @abstractmethod
    def delete(self, key: str):
        ...

def importlib_create_instance(class_full_path: str, constructor_params: dict, setup_values: dict) -> ObjectAsKeyValuePersistenceMixin:
    """
    Creates a persistable class instance given a persisted value and executes its setup_by_dict method

    see :func:`~eric_sse.persistence.ObjectAsKeyValuePersistenceMixin.setup_by_dict`
    """
    path_parts = class_full_path.split('.')
    module = '.'.join(path_parts[:-1])
    klass = path_parts[-1]

    module_object = import_module(module)
    obj: ObjectAsKeyValuePersistenceMixin = getattr(module_object, klass)(**constructor_params)

    try:
        assert isinstance(obj, ObjectAsKeyValuePersistenceMixin)
    except AssertionError as e:
        raise RepositoryError(e) from e

    obj.kv_setup_by_dict(setup_values)

    return obj





from dataclasses import dataclass

import eric_sse
from eric_sse.listener import MessageQueueListener, PersistableListener
from eric_sse.queues import Queue, PersistableQueue

@dataclass
class Connection:
    """
    A connection is just a listener and its related message queue

    :param ~eric_sse.listener.MessageQueueListener listener:
    :param ~eric_sse.queues.Queue queue:
    """
    listener: MessageQueueListener
    queue: Queue
    id: str = eric_sse.generate_uuid()


class PersistableConnection(Connection):
    listener: PersistableListener
    queue: PersistableQueue



import asyncio
from concurrent.futures import ThreadPoolExecutor, Executor
from typing import Callable, AsyncIterable
from eric_sse import get_logger
from eric_sse.entities import AbstractChannel
from eric_sse.repository import PersistableChannel
from eric_sse.listener import MessageQueueListener
from eric_sse.message import SignedMessage, MessageContract
from eric_sse.exception import NoMessagesException
from eric_sse.queues import InMemoryQueue

logger = get_logger()


class SSEChannel(PersistableChannel):
    """
    SSE streaming channel.
    See `Mozilla docs <https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#event_stream_format>`_

    Currently, 'id' field is not supported.
    """

    def __init__(
            self,
            stream_delay_seconds: int = 0,
            retry_timeout_milliseconds: int = 5,
            channel_id: str | None = None
    ):
        super().__init__(stream_delay_seconds=stream_delay_seconds, channel_id=channel_id)
        self.retry_timeout_milliseconds = retry_timeout_milliseconds


    @property
    def kv_key(self) -> str:
        return self.id

    @property
    def kv_setup_values_as_dict(self) -> dict:
        return {
            'stream_delay_seconds': self.stream_delay_seconds,
            'retry_timeout_milliseconds': self.retry_timeout_milliseconds,
            'channel_id': self.id
        }

    @property
    def kv_constructor_params_as_dict(self) -> dict:
        return {
            'stream_delay_seconds': self.stream_delay_seconds,
            'retry_timeout_milliseconds': self.retry_timeout_milliseconds,
            'channel_id': self.id,
        }

    def kv_setup_by_dict(self, setup: dict):
        self.stream_delay_seconds = setup['stream_delay_seconds']
        self.retry_timeout_milliseconds = setup['retry_timeout_milliseconds']

    def adapt(self, msg: MessageContract) -> dict:
        """
        SSE adapter.

        Returns::

            {
                "event": "message type",
                "retry": "channel time out",
                "data": "original payload"
            }
        """
        return {
            "event": msg.type,
            "retry": self.retry_timeout_milliseconds,
            "data": msg.payload
        }


class DataProcessingChannel(AbstractChannel):
    """
    Channel intended for concurrent processing of data.

    Relies on `concurrent.futures.Executor <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.Executor>`_.
    Just override **adapt** method to control output returned to clients

    MESSAGE_TYPE_CLOSED type is intended as end of stream. It should be considered as a reserved Message type.  
    """

    def __init__(self, max_workers: int, stream_delay_seconds: int = 0, executor_class: Executor.__class__ = ThreadPoolExecutor):
        """
        :param max_workers: Num of workers to use
        :param stream_delay_seconds: Can be used to limit response rate of streaming. Only applies to message_stream calls.
        :param executor_class: The constructor of some Executor class. Defaults to  `ThreadPoolExecutor  <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`_.
        """
        super().__init__(stream_delay_seconds=stream_delay_seconds)
        self.max_workers = max_workers
        self.executor_class = executor_class

    async def process_queue(self, listener: MessageQueueListener) -> AsyncIterable[dict]:
        """Performs queue processing of a given listener, returns an AsyncIterable of dictionaries containing message process result. See **adapt** method"""

        with self.executor_class(max_workers=self.max_workers) as e:
            there_are_pending_messages = True
            tasks = []
            loop = asyncio.get_running_loop()
            while there_are_pending_messages:
                try:
                    msg = self._get_queue(listener_id=listener.id).pop()
                    tasks.append(loop.run_in_executor(e, DataProcessingChannel._invoke_callback_and_return, listener.on_message, msg))

                except NoMessagesException:
                    there_are_pending_messages = False

            for task_dome in asyncio.as_completed(tasks):
                task_result = await task_dome
                yield self.adapt(task_result)

    @staticmethod
    def _invoke_callback_and_return(callback: Callable[[MessageContract], None], msg: MessageContract):
        callback(msg)
        return msg

    def adapt(self, msg: MessageContract) -> dict:
        """

        Returns a dictionary in the following format::

            {
                "event": message type
                "data": message payload
            }
        """

        return {
            "event": msg.type,
            "data": msg.payload
        }

class SimpleDistributedApplicationListener(MessageQueueListener):
    """Listener for distributed applications"""

    def __init__(self):
        super().__init__()
        self.__channel: AbstractChannel | None = None
        self.__actions: dict[str, Callable[[MessageContract], list[MessageContract]]] = dict()
        self.__internal_actions: dict[str, Callable[[], None]] = {
            'start': self.start,
            'stop': self.stop
        }

    def set_channel(self, channel: AbstractChannel):
        self.__channel = channel

    def set_action(self, name: str, action: Callable[[MessageContract], list[MessageContract]]):
        """
        Hooks a callable to a string key.

        Callables are selected when listener processes the message depending on its type.

        They should return a list of MessageContract instances corresponding to response to action requested.

        Reserved actions are 'start', 'stop'.
        Receiving a message with one of these types will fire corresponding action.

        """
        if action in self.__internal_actions:
            raise KeyError(f'Trying to set an internal action {action}')
        self.__actions[name] = action

    def dispatch_to(self, receiver: MessageQueueListener, msg: MessageContract):
        signed_message = SignedMessage(sender_id=self.id, msg_type=msg.type, msg_payload=msg.payload)
        self.__channel.dispatch(receiver.id, signed_message)

    def on_message(self, msg: SignedMessage) -> None:
        """Executes action corresponding to message's type"""
        try:
            try:
                self.__internal_actions[msg.type]()
                return
            except KeyError:
                pass

            msgs = self.__actions[msg.type](msg)
            for response in msgs:
                signed_response = SignedMessage(sender_id=self.id, msg_type=response.type, msg_payload=response.payload)
                self.__channel.dispatch(msg.sender_id, signed_response)
        except KeyError:
            logger.debug(f'Unknown action {msg.type}')

class SimpleDistributedApplicationChannel(SSEChannel):

    def register_listener(self, listener: SimpleDistributedApplicationListener):
        listener.set_channel(self)
        super().register_connection(listener, InMemoryQueue())


from abc import ABC
from typing import Iterable

from eric_sse.entities import AbstractChannel
from eric_sse.interfaces import ChannelRepositoryInterface
from eric_sse.persistence import ObjectAsKeyValuePersistenceMixin, KvStorageEngine

class PersistableChannel(AbstractChannel, ObjectAsKeyValuePersistenceMixin, ABC):
    ...


class ChannelRepository(ChannelRepositoryInterface):
    """
    If your concrete repository is a serializable object, you can use this class if correspondant storage engine supports its format

    see :class:`~eric_sse.inmemory.InMemoryChannelRepository`
    """
    def __init__(self, storage_engine: KvStorageEngine):
        self.__storage_engine = storage_engine

    def load_all(self) -> Iterable[PersistableChannel]:
        for obj in self.__storage_engine.fetch_all():
            yield obj

    def load_one(self, key: str) -> PersistableChannel:
        return self.__storage_engine.fetch_one(key)

    def persist(self, persistable: PersistableChannel):
        self.__storage_engine.upsert(
            persistable.kv_key, persistable.kv_setup_values_as_dict
        )

    def delete(self, key: str):
        self.__storage_engine.delete(key)



import time

from eric_sse.listener import MessageQueueListener, PersistableListener
from eric_sse.message import MessageContract
from eric_sse.prefabs import DataProcessingChannel
from eric_sse import get_logger
from eric_sse.queues import InMemoryQueue

logger = get_logger()

class ListenerWrapper(PersistableListener):
    """Wraps a listener to profile its on_message method."""

    def __init__(self, listener: MessageQueueListener, profile_messages: bool = False):
        super().__init__()
        self.listener = listener
        self.profile_messages = profile_messages

    def on_message(self, msg: MessageContract) -> None:
        """Performs on_message profiling"""
        start = time.time()
        self.listener.on_message(msg)

        if self.profile_messages:
            logger.info(f"[PROFILER][MESSAGE] processing time: {time.time() - start}")


class DataProcessingChannelProfiler:

    def __init__(self, channel: DataProcessingChannel):
        """Wraps a channel to profile its process_queue method."""
        self.channel = channel

    def add_listener(self, listener: MessageQueueListener) -> ListenerWrapper:
        """Adds a listener to the channel after having wrapped it"""
        wrapper = ListenerWrapper(listener)
        self.channel.register_connection(wrapper, InMemoryQueue())
        return wrapper

    async def run(self, listener: ListenerWrapper):
        """Runs profile"""
        start = time.time()
        async for _ in self.channel.process_queue(listener):
            pass
        logger.info(f"[PROFILER][QUEUE] processing time: {time.time() - start}")

